@@ -1,6 +1,7 @@
 use std::{marker::PhantomData, sync::Arc};
 
 use futures::{future::poll_fn, pin_mut};
+use once_cell::sync::Lazy;
 use tracing::{error, info, trace};
 
 use crate::{self as elfo};
@@ -12,7 +13,7 @@ use crate::{
     address_book::AddressBook,
     config::AnyConfig,
     demux::Demux,
-    dumping::{self, Direction, Dumper},
+    dumping::{self, Direction, Dump, Dumper, INTERNAL_CLASS},
     envelope::{Envelope, MessageKind},
     errors::{RequestError, SendError, TryRecvError, TrySendError},
     mailbox::RecvResult,
@@ -28,10 +29,11 @@ use self::stats::Stats;
 
 mod stats;
 
+static DUMPER: Lazy<Dumper> = Lazy::new(|| Dumper::new(INTERNAL_CLASS));
+
 /// An actor execution context.
 pub struct Context<C = (), K = Singleton, S = ()> {
     book: AddressBook,
-    dumper: Dumper,
     addr: Addr,
     group: Addr,
     demux: Demux,
@@ -88,7 +90,6 @@ impl<C, K, S> Context<C, K, S> {
     pub fn with<S1>(self, source: S1) -> Context<C, K, Combined<S, S1>> {
         Context {
             book: self.book,
-            dumper: self.dumper,
             addr: self.addr,
             group: self.group,
             demux: self.demux,
@@ -133,7 +134,7 @@ impl<C, K, S> Context<C, K, S> {
     /// ctx.send(SomethingHappened).await?;
     ///
     /// // Fire or log.
-    /// if let Some(err) = ctx.send(SomethingHappened).await {
+    /// if let Ok(err) = ctx.send(SomethingHappened).await {
     ///     warn!("...", error = err);
     /// }
     /// ```
@@ -142,6 +143,81 @@ impl<C, K, S> Context<C, K, S> {
         self.do_send(message, kind).await
     }
 
+    /// Tries to send a message using the routing system.
+    ///
+    /// Returns
+    /// * `Ok(())` if the message has been added to any mailbox.
+    /// * `Err(Full(_))` if some mailboxes are full.
+    /// * `Err(Closed(_))` otherwise.
+    ///
+    /// # Example
+    /// ```ignore
+    /// // Fire and forget.
+    /// let _ = ctx.try_send(SomethingHappened);
+    ///
+    /// // Fire or fail.
+    /// ctx.try_send(SomethingHappened)?;
+    ///
+    /// // Fire or log.
+    /// if let Err(err) = ctx.try_send(SomethingHappened) {
+    ///     warn!("...", error = err);
+    /// }
+    /// ```
+    pub fn try_send<M: Message>(&self, message: M) -> Result<(), TrySendError<M>> {
+        let kind = MessageKind::Regular { sender: self.addr };
+
+        // XXX: unify with `do_send`.
+        self.stats.sent_messages_total::<M>();
+
+        trace!("> {:?}", message);
+        if let Some(permit) = DUMPER.acquire() {
+            permit.record(Dump::message(message.clone(), &kind, Direction::Out));
+        }
+
+        let envelope = Envelope::new(message, kind).upcast();
+        let addrs = self.demux.filter(&envelope);
+
+        if addrs.is_empty() {
+            return Err(TrySendError::Closed(e2m(envelope)));
+        }
+
+        if addrs.len() == 1 {
+            return match self.book.get(addrs[0]) {
+                Some(object) => object.try_send(envelope).map_err(|err| err.map(e2m)),
+                None => Err(TrySendError::Closed(e2m(envelope))),
+            };
+        }
+
+        let mut unused = None;
+        let mut has_full = false;
+        let mut success = false;
+
+        // TODO: use the visitor pattern in order to avoid extra cloning.
+        for addr in addrs {
+            let envelope = unused.take().or_else(|| envelope.duplicate(&self.book));
+            let envelope = ward!(envelope, break);
+
+            match self.book.get(addr) {
+                Some(object) => match object.try_send(envelope) {
+                    Ok(()) => success = true,
+                    Err(err) => {
+                        has_full |= err.is_full();
+                        unused = Some(err.into_inner());
+                    }
+                },
+                None => unused = Some(envelope),
+            };
+        }
+
+        if success {
+            Ok(())
+        } else if has_full {
+            Err(TrySendError::Full(e2m(envelope)))
+        } else {
+            Err(TrySendError::Closed(e2m(envelope)))
+        }
+    }
+
     /// Returns a request builder.
     ///
     /// # Example
@@ -184,15 +260,15 @@ impl<C, K, S> Context<C, K, S> {
         self.stats.sent_messages_total::<M>();
 
         trace!("> {:?}", message);
-        if self.dumper.is_enabled() {
-            self.dumper.dump_message(&message, &kind, Direction::Out);
+        if let Some(permit) = DUMPER.acquire() {
+            permit.record(Dump::message(message.clone(), &kind, Direction::Out));
         }
 
         let envelope = Envelope::new(message, kind).upcast();
         let addrs = self.demux.filter(&envelope);
 
         if addrs.is_empty() {
-            return Err(SendError(envelope.do_downcast().into_message()));
+            return Err(SendError(e2m(envelope)));
         }
 
         if addrs.len() == 1 {
@@ -200,8 +276,8 @@ impl<C, K, S> Context<C, K, S> {
                 Some(object) => object
                     .send(self, envelope)
                     .await
-                    .map_err(|err| SendError(err.0.do_downcast().into_message())),
-                None => Err(SendError(envelope.do_downcast().into_message())),
+                    .map_err(|err| SendError(e2m(err.0))),
+                None => Err(SendError(e2m(envelope))),
             };
         }
 
@@ -228,7 +304,7 @@ impl<C, K, S> Context<C, K, S> {
         if success {
             Ok(())
         } else {
-            Err(SendError(envelope.do_downcast().into_message()))
+            Err(SendError(e2m(envelope)))
         }
     }
 
@@ -267,8 +343,8 @@ impl<C, K, S> Context<C, K, S> {
         self.stats.sent_messages_total::<M>();
 
         trace!(to = %recipient, "> {:?}", message);
-        if self.dumper.is_enabled() {
-            self.dumper.dump_message(&message, &kind, Direction::Out);
+        if let Some(permit) = DUMPER.acquire() {
+            permit.record(Dump::message(message.clone(), &kind, Direction::Out));
         }
 
         let entry = self.book.get_owned(recipient);
@@ -276,7 +352,7 @@ impl<C, K, S> Context<C, K, S> {
         let envelope = Envelope::new(message, kind);
         let fut = object.send(self, envelope.upcast());
         let result = fut.await;
-        result.map_err(|err| SendError(err.0.do_downcast().into_message()))
+        result.map_err(|err| SendError(e2m(err.0)))
     }
 
     /// Tries to send a message to the specified recipient.
@@ -306,22 +382,17 @@ impl<C, K, S> Context<C, K, S> {
         let kind = MessageKind::Regular { sender: self.addr };
 
         trace!(to = %recipient, "> {:?}", message);
-        if self.dumper.is_enabled() {
-            self.dumper.dump_message(&message, &kind, Direction::Out);
+        if let Some(permit) = DUMPER.acquire() {
+            permit.record(Dump::message(message.clone(), &kind, Direction::Out));
         }
 
-        let entry = self.book.get_owned(recipient);
+        let entry = self.book.get(recipient);
         let object = ward!(entry, return Err(TrySendError::Closed(message)));
         let envelope = Envelope::new(message, kind);
 
-        object.try_send(envelope.upcast()).map_err(|err| match err {
-            TrySendError::Full(envelope) => {
-                TrySendError::Full(envelope.do_downcast().into_message())
-            }
-            TrySendError::Closed(envelope) => {
-                TrySendError::Closed(envelope.do_downcast().into_message())
-            }
-        })
+        object
+            .try_send(envelope.upcast())
+            .map_err(|err| err.map(e2m))
     }
 
     /// Responds to the requester with the provided response.
@@ -343,15 +414,18 @@ impl<C, K, S> Context<C, K, S> {
         self.stats.sent_messages_total::<R::Wrapper>();
 
         let sender = token.sender;
+        let message = R::Wrapper::from(message);
+        let kind = MessageKind::Response {
+            sender: self.addr(),
+            request_id: token.request_id,
+        };
 
         trace!(to = %sender, "> {:?}", message);
-        if self.dumper.is_enabled() {
-            self.dumper
-                .dump_response::<R>(&message, token.request_id, Direction::Out);
+        if let Some(permit) = DUMPER.acquire() {
+            permit.record(Dump::message(message.clone(), &kind, Direction::Out));
         }
 
-        let message = R::Wrapper::from(message);
-        let envelope = Envelope::new(message, MessageKind::Regular { sender }).upcast();
+        let envelope = Envelope::new(message, kind).upcast();
         let object = ward!(self.book.get(token.sender));
         let actor = ward!(object.as_actor());
         actor
@@ -499,15 +573,18 @@ impl<C, K, S> Context<C, K, S> {
         let message = envelope.message();
         trace!("< {:?}", message);
 
-        if self.dumper.is_enabled() {
-            self.dumper.dump(
-                dumping::Direction::In,
-                "",
-                message.name(),
-                message.protocol(),
-                dumping::MessageKind::from_message_kind(envelope.message_kind()),
-                message.erase(),
-            )
+        if let Some(permit) = DUMPER.acquire() {
+            // TODO: reuse `Dump::message`, it requires `AnyMessage: Message`.
+            let dump = Dump::builder()
+                .direction(Direction::In)
+                .message_name(message.name())
+                .message_protocol(message.protocol())
+                .message_kind(dumping::MessageKind::from_message_kind(
+                    envelope.message_kind(),
+                ))
+                .do_finish(message.erase());
+
+            permit.record(dump);
         }
 
         // We should change the status after dumping the original message
@@ -556,7 +633,6 @@ impl<C, K, S> Context<C, K, S> {
     pub fn pruned(&self) -> Context {
         Context {
             book: self.book.clone(),
-            dumper: self.dumper.clone(),
             addr: self.addr,
             group: self.group,
             demux: self.demux.clone(),
@@ -572,14 +648,9 @@ impl<C, K, S> Context<C, K, S> {
         &self.book
     }
 
-    pub(crate) fn dumper(&self) -> &Dumper {
-        &self.dumper
-    }
-
     pub(crate) fn with_config<C1>(self, config: Arc<C1>) -> Context<C1, K, S> {
         Context {
             book: self.book,
-            dumper: self.dumper,
             addr: self.addr,
             group: self.group,
             demux: self.demux,
@@ -604,7 +675,6 @@ impl<C, K, S> Context<C, K, S> {
     pub(crate) fn with_key<K1>(self, key: K1) -> Context<C, K1, S> {
         Context {
             book: self.book,
-            dumper: self.dumper,
             addr: self.addr,
             group: self.group,
             demux: self.demux,
@@ -617,6 +687,10 @@ impl<C, K, S> Context<C, K, S> {
     }
 }
 
+fn e2m<M: Message>(envelope: Envelope) -> M {
+    envelope.do_downcast::<M>().into_message()
+}
+
 #[cold]
 fn on_first_recv(stage: &mut Stage, actor: &Actor) {
     if actor.is_initializing() {
@@ -641,10 +715,9 @@ fn on_recv_after_close() {
 }
 
 impl Context {
-    pub(crate) fn new(book: AddressBook, dumper: Dumper, demux: Demux) -> Self {
+    pub(crate) fn new(book: AddressBook, demux: Demux) -> Self {
         Self {
             book,
-            dumper,
             addr: Addr::NULL,
             group: Addr::NULL,
             demux,
@@ -661,7 +734,6 @@ impl<C, K: Clone> Clone for Context<C, K> {
     fn clone(&self) -> Self {
         Self {
             book: self.book.clone(),
-            dumper: self.dumper.clone(),
             addr: self.addr,
             group: self.group,
             demux: self.demux.clone(),
@@ -760,15 +832,18 @@ impl<'c, C: 'static, K, S, R: Request> RequestBuilder<'c, C, S, K, R, Any> {
 
         let mut data = actor.request_table().wait(request_id).await;
         if let Some(Some(envelope)) = data.pop() {
+            let envelope = envelope.do_downcast::<R::Wrapper>();
+
             // TODO: increase a counter.
-            let message = envelope.do_downcast::<R::Wrapper>().into_message().into();
-            trace!("< {:?}", message);
-            if self.context.dumper.is_enabled() {
-                self.context
-                    .dumper
-                    .dump_response::<R>(&message, request_id, Direction::In);
+            trace!("< {:?}", envelope.message());
+            if let Some(permit) = DUMPER.acquire() {
+                permit.record(Dump::message(
+                    envelope.message().clone(),
+                    envelope.message_kind(),
+                    Direction::In,
+                ));
             }
-            Ok(message)
+            Ok(envelope.into_message().into())
         } else {
             // TODO: should we dump it and increase a counter?
             Err(RequestError::Ignored)
@@ -799,36 +874,32 @@ impl<'c, C: 'static, K, S, R: Request> RequestBuilder<'c, C, K, S, R, All> {
             return vec![Err(RequestError::Closed(err.0))];
         }
 
-        let responses = actor
+        actor
             .request_table()
             .wait(request_id)
             .await
             .into_iter()
             .map(|opt| match opt {
-                Some(envelope) => Ok(envelope.do_downcast::<R::Wrapper>().into_message().into()),
+                Some(envelope) => Ok(envelope.do_downcast::<R::Wrapper>()),
                 None => Err(RequestError::Ignored),
             })
-            .inspect(|res| {
+            .map(|res| {
+                let envelope = res?;
+
                 // TODO: increase a counter.
-                if let Ok(message) = res {
-                    trace!("< {:?}", message);
+                trace!("< {:?}", envelope.message());
+
+                // TODO: `acquire_many` or even unconditionally?
+                if let Some(permit) = DUMPER.acquire() {
+                    permit.record(Dump::message(
+                        envelope.message().clone(),
+                        envelope.message_kind(),
+                        Direction::In,
+                    ));
                 }
+                Ok(envelope.into_message().into())
             })
-            .collect::<Vec<_>>();
-
-        if self.context.dumper.is_enabled() {
-            #[allow(clippy::manual_flatten)]
-            for response in &responses {
-                // TODO: dump errors.
-                if let Ok(res) = response {
-                    self.context
-                        .dumper
-                        .dump_response::<R>(res, request_id, Direction::In);
-                }
-            }
-        }
-
-        responses
+            .collect()
     }
 }
 
